//ODEHuman.cpp
//--------------------** Description ** -----------------------------------------------
// a set of joints and rigid boxes representing a human figure
// May be posed using a Posture and the state of the simulation can be returned as a Posture
//-------------------------------------------------------------------------------------
//Copyright 2005 Michael J Mandel
//Licensed under the LGPL (see license.txt for more information)

#include "ODEHuman.h"
#include "ODESimulation.h"
#include "ODEBox.h"
#include "glut.h"

//cap the torque generated by the controller to a reasonable value
#define MAX_TORQUE 60.0

ODEHuman::ODEHuman(const ODEPosition &humanPos, ODESimulation *sim, bool hasBody) : ODEObject(sim, hasBody), mEnabled(false),  mControllerEnabled(false)
{

	SetControllerProperties(controlPropsDefault);

	CreateBodiesAndJoints(humanPos);

	SetEnabled(true);
}

ODEHuman::~ODEHuman()
{
	//remove all bodies from the simulation
	std::vector<ODEBox*>::iterator it;
	for (it = rigidLinks.begin() ; it != rigidLinks.end() ; it++)
	{
		delete (*it); //ODEBox destructor will take care of the rest
	}

	mBodyID = 0;
	mGeomID = 0;
}

//creates all the joints and bodies for a human in a reference, or base posture
void ODEHuman::CreateBodiesAndJoints(const ODEPosition &humanPos, bool allowAttachToWorld)
{
	rigidLinks.resize(NUM_BODY_PARTS);
	rigidJoints.resize(NUM_BODY_PARTS);

	//create all the bodies
	for(int i = 0 ; i < NUM_BODY_PARTS ; i++)
	{
		ODEBox *newBox = new ODEBox(bodyProps[i].size, mSimulator, true);
		rigidLinks[i] = newBox;
		newBox->SetID(i);
		newBox->SetMass(bodyProps[i].mass);

		if(i == lowerback)
		{
			mBodyID = newBox->GetBodyID();
			mGeomID = newBox->GetGeomID();
		}

		newBox->SetPos(bodyProps[i].pos);
		newBox->SetRot(bodyProps[i].rot);

		//bodies will be enabled when joints are created
		newBox->SetEnabled(false);
	}


	//create joints
	dJointID newJoint = 0;
	for(i = 0 ; i < NUM_BODY_PARTS ; i++)
	{
		odeJointProperty curProp = jointProps[i];

		if(bodyProps[curProp.attachedBody].enabled )
		{
			if(curProp.attachToWorld && !allowAttachToWorld)
				continue;

			switch(curProp.jointType)
			{
				case JOINT_TYPE_HINGE:

					newJoint = dJointCreateHinge(mSimulator->GetWorldID(), 0);
					rigidJoints[curProp.attachedBody] = newJoint;
					if(curProp.attachToWorld)
						dJointAttach(newJoint, rigidLinks[curProp.attachedBody]->GetBodyID(), 0);
					else
						dJointAttach(newJoint, rigidLinks[curProp.attachedBody]->GetBodyID(), rigidLinks[curProp.parentBody]->GetBodyID());

					dJointSetHingeAnchor(newJoint, WORLD_TO_SIM(curProp.jointAnchor[0]), WORLD_TO_SIM(curProp.jointAnchor[1]), WORLD_TO_SIM(curProp.jointAnchor[2]));
					dJointSetHingeAxis(newJoint, curProp.rotAxis1[0], curProp.rotAxis1[1], curProp.rotAxis1[2]);
 
					//joint limit constraints
					if(!(curProp.attachToWorld))
					{
						//set them twice to make sure it "sticks"
						dJointSetHingeParam (newJoint, dParamLoStop, curProp.loJointRange1);
						dJointSetHingeParam (newJoint, dParamHiStop, curProp.highJointRange1);
						dJointSetHingeParam (newJoint, dParamLoStop, curProp.loJointRange1);
						dJointSetHingeParam (newJoint, dParamHiStop, curProp.highJointRange1);
					}
				break;

				case JOINT_TYPE_UNIVERSAL:

					newJoint = dJointCreateUniversal(mSimulator->GetWorldID(), 0);
					rigidJoints[curProp.attachedBody] = newJoint;
					if(curProp.attachToWorld)
						dJointAttach(newJoint, rigidLinks[curProp.attachedBody]->GetBodyID(), 0);
					else
						dJointAttach(newJoint, rigidLinks[curProp.attachedBody]->GetBodyID(), rigidLinks[curProp.parentBody]->GetBodyID());

					dJointSetUniversalAnchor(newJoint, WORLD_TO_SIM(curProp.jointAnchor[0]), WORLD_TO_SIM(curProp.jointAnchor[1]), WORLD_TO_SIM(curProp.jointAnchor[2]));


					dJointSetUniversalAxis1(newJoint, curProp.rotAxis1[0], curProp.rotAxis1[1], curProp.rotAxis1[2]); 
					dJointSetUniversalAxis2(newJoint, curProp.rotAxis2[0], curProp.rotAxis2[1], curProp.rotAxis2[2]);

					//joint limit constraints
					if(!(curProp.attachToWorld))
					{
						//set them twice to make sure it "sticks"
						dJointSetUniversalParam (newJoint, dParamLoStop, curProp.loJointRange1);
						dJointSetUniversalParam (newJoint, dParamHiStop, curProp.highJointRange1);
						dJointSetUniversalParam (newJoint, dParamLoStop, curProp.loJointRange1);
						dJointSetUniversalParam (newJoint, dParamHiStop, curProp.highJointRange1);
						dJointSetUniversalParam (newJoint, dParamLoStop2, curProp.loJointRange2);
						dJointSetUniversalParam (newJoint, dParamHiStop2, curProp.highJointRange2);
						dJointSetUniversalParam (newJoint, dParamLoStop2, curProp.loJointRange2);
						dJointSetUniversalParam (newJoint, dParamHiStop2, curProp.highJointRange2);
					}
				break;
				
				default:
				break;
			}

			rigidLinks[i]->SetEnabled(true);
		}

	}
	

	//disable collisions between hips and upper arms since they are
	//not connected, but may be touching
	rigidLinks[lhipjoint]->DisableCollisionsWith(rigidLinks[rhipjoint]);
	rigidLinks[lclavicle]->DisableCollisionsWith(rigidLinks[rclavicle]);
}

//add a force to a particular body
void ODEHuman::AddForceToBody(int bodyID, const ODEVector &force)
{
	ODEBox *body = GetBody(bodyID);

	if(body)
	{
		body->AddForce(force);
	}
}

//enable/disable all bodies and joints
void ODEHuman::SetEnabled(bool b)
{
	mEnabled = b;

	//loop through all bodies and enable them
	for(int i = 0 ; i < NUM_BODY_PARTS ; i++)
	{
		odeJointProperty curProp = jointProps[i];

		//enable appropriate bodies
		if(mEnabled && bodyProps[curProp.attachedBody].enabled)
		{
			rigidLinks[curProp.attachedBody]->SetEnabled(true);

			if(!curProp.attachToWorld)
				rigidLinks[curProp.parentBody]->SetEnabled(true);
		}
		else if(!mEnabled)
		{
			rigidLinks[curProp.attachedBody]->SetEnabled(false);
		}

	}
}



void ODEHuman::Render()
{
	//individual bodies will be rendered automatically because they are ODEObjects....
}

void ClampTorque(float &torque)
{
	if(torque > MAX_TORQUE)
		torque = MAX_TORQUE;
	if(torque < -MAX_TORQUE)
		torque = -MAX_TORQUE;
}

//compute and apply PD-controller torques to each body part
void ODEHuman::applyControlTorques()
{
	if(mControllerEnabled)
	{	
		for(int i = 0 ; i < NUM_BODY_PARTS ; i++)
		{
			odeControlProperty curControlProp = mCurControllerProps[i];
			odeJointProperty curJointProp = jointProps[i];

			if(curControlProp.isControllable && bodyProps[curJointProp.attachedBody].enabled && !curJointProp.attachToWorld)
			{

				float curAngleX, curVelX, curAngleY, curVelY;
				float torqueX, torqueY;
				torqueX = torqueY = 0.0f;

				//desired joint angles are the base pose from the initial setup...
				float desiredX = 0;
				float desiredY = 0;

				// Apply the torques
				switch(curJointProp.jointType)
				{
					case JOINT_TYPE_HINGE:
						//get current angle
						curAngleX = dJointGetHingeAngle(rigidJoints[curJointProp.attachedBody]);
						//get current speed
						curVelX = dJointGetHingeAngleRate (rigidJoints[curJointProp.attachedBody]);
						//compute torque
						torqueX = curControlProp.ks1*(desiredX - curAngleX) - curControlProp.kd1*curVelX;

						//apply maximum torque constraint
						ClampTorque(torqueX);

						dJointAddHingeTorque(rigidJoints[curJointProp.attachedBody], torqueX);
					break;

					case JOINT_TYPE_UNIVERSAL:
						//get current angle
						curAngleX = dJointGetUniversalAngle1(rigidJoints[curJointProp.attachedBody]);
						curAngleY = dJointGetUniversalAngle2(rigidJoints[curJointProp.attachedBody]);
						//get current speed
						curVelX = dJointGetUniversalAngle1Rate (rigidJoints[curJointProp.attachedBody]);
						curVelY = dJointGetUniversalAngle2Rate (rigidJoints[curJointProp.attachedBody]);
						//compute torque
						torqueX = curControlProp.ks1*(desiredX - curAngleX) - curControlProp.kd1*curVelX;
						torqueY = curControlProp.ks2*(desiredY - curAngleY) - curControlProp.kd2*curVelY;

						//apply maximum torque constraint
						ClampTorque(torqueX);
						ClampTorque(torqueY);

						dJointAddUniversalTorques(rigidJoints[curJointProp.attachedBody], torqueX,torqueY);
					break;
					
					default:

					break;
				}
			}
		}
	}
}

//set the controller gains used by the controller
void ODEHuman::SetControllerProperties(odeControlProperty *props)
{
	if(props)
		mCurControllerProps = props;
}

//activates/deactivates the PD-controller
void ODEHuman::EnableController(bool b) 
{ 
	if(b != mControllerEnabled)
		mControllerEnabled = b; 
}

bool ODEHuman::Update()
{
	//inidividual bodies will be updated automatically because they are ODEObjects

	//just need to apply the controller torques...
	applyControlTorques();

	return true;
}