/*
SimpleControllerDemo: 
This demo sets up a ragdoll character using the Open Dynamics Engine.
You can then enable a simple proportional derivative controller that
compute torques at each joint to drive the body into a desired pose...
Copyright (C) 2005  Michael J Mandel  
Contact me at mmandel@gmail.com with any questions/comments
Licensed under the LGPL (see license.txt for more information)
*/


#include <string.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
 
#include "glut.h"
#include "ODEObject.h"
#include "ODEBox.h"
#include "ODESimulation.h"
#include "ODEHuman.h"
#include "WinGLFont.h"
#include "TimeStamp.h"

ODESimulation *gSimulator;
ODEHuman *gRagdoll;

struct Camera {
  double zoom;
  //euler angle rotation
  double rx;
  double ry;
  double rz;
  //translation
  double tx;
  double ty;
  double tz;
};

Camera cam;

TimeStamp worldTimer;

WinGLFont *glText;
int gWidth, gHeight;

void ApplyCameraTransform()
{
    glTranslated(cam.tx, cam.ty, cam.tz);

    glRotated(cam.rx, 1.0, 0.0, 0.0);
    glRotated(cam.ry, 0.0, 1.0, 0.0); 
	glRotated(cam.rz, 0.0, 0.0, 1.0);  
    
	glScaled(cam.zoom, cam.zoom, cam.zoom);
}

void init (void)
{
	glShadeModel(GL_SMOOTH);							
	glClearColor(0.0f, 0.0f, 0.0f, 0.5f);				
	glClearDepth(1.0f);									
	glEnable(GL_DEPTH_TEST);							
	glDepthFunc(GL_LEQUAL);								
	glEnable ( GL_COLOR_MATERIAL );
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

	cam.zoom = 1;
	cam.rx = 13;
	cam.ry = 0;
    cam.rz = 0;    
    cam.tx = 0;
    cam.ty = 0;
    cam.tz = 0;

	//set up some lights
    GLfloat white[] = {1., 1., 1., 1.};
	GLfloat light0_position[] = {1., 1., 5., 0.};
    GLfloat light1_position[] = {-3., 1., -1., 0.};
    GLfloat light2_position[] = {1., 8., -2., 0.};
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, white);	
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, white);
    glLightfv(GL_LIGHT0, GL_POSITION, light0_position);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, white);
    glEnable(GL_LIGHT0);
    glLightfv(GL_LIGHT1, GL_POSITION, light1_position);
    glLightfv(GL_LIGHT1, GL_DIFFUSE, white);
    glEnable(GL_LIGHT1);
    glLightfv(GL_LIGHT2, GL_POSITION, light2_position);
    glLightfv(GL_LIGHT2, GL_DIFFUSE, white);
    glEnable(GL_LIGHT2);
    glEnable(GL_NORMALIZE);
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);	
}

void reshape (int w, int h)
{
  gWidth = w;
  gHeight = h;

  glViewport     ( 0, 0, w, h );
  glMatrixMode   ( GL_PROJECTION );  
  glLoadIdentity ( );                
  if ( h==0 )
     gluPerspective ( 80, ( float ) w, .50, 1000.0 );
  else
     gluPerspective ( 80, ( float ) w / ( float ) h, .50, 1000.0 );
  glMatrixMode   ( GL_MODELVIEW ); 
  glLoadIdentity ( );   
}

//primary draw function
void display()
{
	glClearColor(1,1,1,1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();
	glTranslatef(0,5.0f,-65.0f);

	ApplyCameraTransform();

	//step the simulation and render the results
	//a constant time step is more stable
	gSimulator->Update(1.0f / 50.0f);
	gSimulator->Render();

	//on screen instructions
	glColor3f(0, 0, 0);

	//include some status text
	if(gRagdoll->isControllerEnabled())
		glText->glPrintf(5, 10,"Controller Enabled (press 'c' to disable)");
	else
		glText->glPrintf(5, 10,"Controller NOT Enabled (press 'c')");
	glText->glPrintf(5,10 + 1*glText->fontSize(), "Press 'f' to apply force to pelvis...");
	glText->glPrintf(5,10 + 2*glText->fontSize(), "Rotate and Zoom Using Right/Left Mouse Button + Movement");

	//lock framerate at 60hz
	while(worldTimer.Elapsed() < (1.0f / 60.0f))
	{  }
	glutSwapBuffers();
	worldTimer.Stamp();
}

//just redisplay constantly to let the console update...
void idle()
{
	glutPostRedisplay();
}

//handle incoming key events and send to console
void key(unsigned char inkey, int px, int py)
{
	switch (inkey)
	{
		case 27:	//escape
			exit ( 0 );  
		break;  
		case 'c': //enable the controller
			gRagdoll->EnableController(!gRagdoll->isControllerEnabled());
		break;
		case 'f': //apply a force to the pelvis
			ODEVector force;
			force[0] = force[1] = 0; force[2] = -1000;
			gRagdoll->AddForceToBody(0, force);
		default:      
			break;
	}

}


static int left_down = 0;
static int middle_down = 0;
static int right_down = 0;
static int prev_x = 0;
static int prev_y = 0;

void mouse_motion(int x, int y)
{
    
	int delta_x= x - prev_x;
    int delta_y= y - prev_y; 

	if ((left_down && !middle_down && !right_down))
	{
		if(abs(delta_x) > abs(delta_y))
			cam.ry += (GLdouble) (delta_x*.35);          
        else
			cam.rx += (GLdouble) (delta_y*.35);	 
	}
	else if((right_down && !middle_down && !left_down))
	{
        if(abs(delta_y) > abs(delta_x))
        {
            cam.zoom -= (GLdouble) delta_y/100.0;
            if(cam.zoom < 0.) cam.zoom = 0;
		}
	}
	else if((middle_down && !left_down && !right_down))
	{
         cam.tx += (GLdouble) (delta_x/10.0);
         cam.ty -= (GLdouble) (delta_y/10.0); 
	}	

	prev_x = x;
	prev_y = y;
}

void mouse (int button, int state, int x, int y)
{
	if(state == GLUT_DOWN)
	{
		prev_x = x;
		prev_y = y;
	}

	if((state == GLUT_DOWN) && (button == GLUT_MIDDLE_BUTTON))
		middle_down = 1;

	if((state == GLUT_UP) && (button == GLUT_MIDDLE_BUTTON))
		middle_down = 0;

	if((state == GLUT_DOWN) && (button == GLUT_RIGHT_BUTTON))
		right_down = 1;

	if((state == GLUT_UP) && (button == GLUT_RIGHT_BUTTON))
		right_down = 0;

	if((state == GLUT_DOWN) && (button == GLUT_LEFT_BUTTON))
		left_down = 1;

	if((state == GLUT_UP) && (button == GLUT_LEFT_BUTTON))
		left_down = 0;
}

int main (int argc, const char * argv[])
{
    glutInit(&argc, (char **)argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
	glutInitWindowPosition (300, 50);
	glutInitWindowSize (640, 480);
	char *title = "Simple PD Controller Demo";
	glutCreateWindow(title);

    init(); // standard GL init

    glutReshapeFunc (reshape);
    glutDisplayFunc (display);
	glutKeyboardFunc (key);
	glutMouseFunc (mouse);
	glutMotionFunc (mouse_motion);
	glutIdleFunc(idle);

	//create the ODE simulator
	gSimulator = new ODESimulation();
	//create ragdoll character
	ODEPosition pos;
	pos[0] = pos[1] = pos[2] = 0.0f;
	gRagdoll = new ODEHuman(pos,gSimulator);

	HWND hWnd = FindWindow("GLUT", title);
	glText = new WinGLFont(hWnd, "Courier", 18);

	worldTimer.Stamp();

    glutMainLoop();

	delete gRagdoll;
	delete gSimulator;

	return 0;
}
